//-----------------------------------------------------------------------------
// Copyright (c) 2000 Andreas Jnsson
//-----------------------------------------------------------------------------
#include "tripatch.h"
#include "vertex.h"

//-----------------------------------------------------------------------------
// CTriPatch()
//-----------------------------------------------------------------------------
CTriPatch::CTriPatch()
{
	m_szWindowTitle = "Triangular Cubic Beizer Patch";

	m_bShowNormals = false;
	m_bShowSurface = true;
	m_bShowControl = false;
}

//-----------------------------------------------------------------------------
// Initialize()
//-----------------------------------------------------------------------------
HRESULT CTriPatch::Initialize(char *szCmdLine)
{
	// Setup lighting
	D3DLIGHT7 Light;
	Light.dltType = D3DLIGHT_POINT;
	Light.dcvDiffuse = D3DXCOLOR(1,1,1,0);
	Light.dvPosition = D3DVECTOR(0,0,-3);
	Light.dvRange = D3DLIGHT_RANGE_MAX;
	Light.dvAttenuation0 = 1;
	Light.dvAttenuation1 = 0;
	Light.dvAttenuation2 = 0;
	m_pD3DDev->LightEnable(0, TRUE);
	m_pD3DDev->SetLight(0, &Light);
	m_pD3DDev->SetRenderState(D3DRENDERSTATE_LIGHTING, TRUE);
	
	// Background color
	m_pD3DX->SetClearColor(D3DRGBA(0.5f,0.0f,0.0f,0));

	// Make sure the normals are unit length
	m_pD3DDev->SetRenderState(D3DRENDERSTATE_NORMALIZENORMALS, TRUE);

	// Setup the projection matrix
	D3DXMATRIX M;
	D3DXMatrixPerspectiveLH(&M, 1, 0.75f, 0.5f, 100);
	m_pD3DDev->SetTransform(D3DTRANSFORMSTATE_PROJECTION, (D3DMATRIX*)&M);

	// Setup the view matrix
	D3DXMatrixTranslation(&M, 0, 0, 3.0f);
	m_pD3DDev->SetTransform(D3DTRANSFORMSTATE_VIEW, (D3DMATRIX*)&M);

	// Initial angle
	m_fAngle = 0;


	
	
	// Allocate memory for vertices and indices
	m_nNumDivs = (1<<4)+1;
	m_nNumVertices = m_nNumDivs * ((m_nNumDivs-1)/2 + 1);
	m_pVertices = new CVertex[m_nNumVertices];


	// Setup the indices
	m_nNumIndices = ((m_nNumDivs-1)*(m_nNumDivs-1)+1)*3;
	m_pIndices = new WORD[m_nNumIndices];
	int Index = 0;
	int i0 = 0, i1 = 0, i;
	for( int n = 0; n < m_nNumDivs; n++ )
	{
		i0 = i1;
		i1 += m_nNumDivs-n;

		for( i = 0; i < m_nNumDivs-2-n; i++ )
		{
			m_pIndices[Index++] = i0+i;
			m_pIndices[Index++] = i0+i+1;
			m_pIndices[Index++] = i1+i;

			m_pIndices[Index++] = i0+i+1;
			m_pIndices[Index++] = i1+i+1;
			m_pIndices[Index++] = i1+i;
		}

		m_pIndices[Index++] = i0+i;
		m_pIndices[Index++] = i0+i+1;
		m_pIndices[Index++] = i1+i;

		if( Index > m_nNumIndices )
			OutputDebugString("");
	}

	return S_OK;
}

//-----------------------------------------------------------------------------
// Update()
//-----------------------------------------------------------------------------
HRESULT CTriPatch::UpdateScene()
{
	m_fAngle = 1.5f*(float)sin(0.5f*m_SceneTimer.GetFrameTime());

	D3DXMATRIX Rot;
	D3DXMatrixRotationY(&Rot, m_fAngle);
	m_pD3DDev->SetTransform(D3DTRANSFORMSTATE_WORLD, (D3DMATRIX*)Rot);

	return S_OK;
}

//-----------------------------------------------------------------------------
// RenderScene()
//-----------------------------------------------------------------------------
HRESULT CTriPatch::RenderScene()
{
	m_pD3DX->Clear(D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER);

	// Begin a new scene
    HRESULT hr = m_pD3DDev->BeginScene();
	if( SUCCEEDED( hr ) )
	{
		RenderBezier();

		// The scene is finished
        hr = m_pD3DDev->EndScene();
	}

	if( FAILED(hr) )
	{
		ReportError(hr, __FILE__, __LINE__);
		return hr;
	}

	return S_OK;
}

//-----------------------------------------------------------------------------
// RenderBezier()
//-----------------------------------------------------------------------------
void CTriPatch::RenderBezier()
{
	// The control points are ordered like this:
	//
	//    3
	//   2 6
	//  1 5 8
	// 0 4 7 9
	CVertex P[10];
	P[0].v = D3DVECTOR( -1.5f,-1, 0);
	P[1].v = D3DVECTOR(-0.75f, 0, 1);
	P[2].v = D3DVECTOR(-0.25f, 1,-1);
	P[3].v = D3DVECTOR(     0, 2, 0);
	P[4].v = D3DVECTOR( -0.5f,-1, 1);
	P[5].v = D3DVECTOR(     0, 0, 0);
	P[6].v = D3DVECTOR( 0.25f, 1, 1);
	P[7].v = D3DVECTOR(  0.5f,-1,-1);
	P[8].v = D3DVECTOR( 0.75f, 0, 1);
	P[9].v = D3DVECTOR(  1.5f,-1, 0);

	// Compute the surface vertices
	if( m_bShowNormals || m_bShowSurface )
		CubicBezier(P);

	// Render the patch
	if( m_bShowSurface )
	{
		m_pD3DDev->DrawIndexedPrimitive(D3DPT_TRIANGLELIST, FVF_VERTEX, 
		                                m_pVertices, m_nNumVertices,
								        m_pIndices, m_nNumIndices, 0);
	}

	// Show the control net
	if( m_bShowControl )
	{
		m_pD3DDev->SetRenderState(D3DRENDERSTATE_LIGHTING, FALSE);

		for( int n = 0; n < 10; n++ )
			P[n].c = D3DRGB(0,1,0);

		WORD Indices[19] = {0,1,2,3,6,2,5,6,8,9,7,8,5,7,4,5,1,4,0};
		m_pD3DDev->DrawIndexedPrimitive(D3DPT_LINESTRIP, FVF_VERTEX, P, 10, Indices, 19, 0);
		m_pD3DDev->SetRenderState(D3DRENDERSTATE_LIGHTING, TRUE);
	}	

	// Show the vertex normals
	if( m_bShowNormals )
	{
		CVertex *pNormals = new CVertex[m_nNumVertices*2];
		for( int n = 0; n < m_nNumVertices*2; n += 2 )
		{
			pNormals[n].v = m_pVertices[n/2].v;
			pNormals[n+1].v = pNormals[n].v + 0.1f*m_pVertices[n/2].n;
			pNormals[n].n = pNormals[n+1].n = m_pVertices[n/2].n;
			pNormals[n].c = pNormals[n+1].c = D3DRGBA(0,0,1,1);
		}

		m_pD3DDev->DrawPrimitive(D3DPT_LINELIST, FVF_VERTEX, pNormals, m_nNumVertices*2, 0);
		delete[] pNormals;
	}
}


//-----------------------------------------------------------------------------
// Terminate()
//-----------------------------------------------------------------------------
HRESULT CTriPatch::Terminate()
{
	// Free memory
	if( m_pVertices )
	{
		delete[] m_pVertices;
		m_pVertices = 0;
	}

	if( m_pIndices )
	{
		delete[] m_pIndices;
		m_pIndices = 0;
	}

	return S_OK;
}






//-----------------------------------------------------------------------------
// MsgProc()
//-----------------------------------------------------------------------------
LRESULT CTriPatch::MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam )
{
	switch( uMsg )
	{
	case WM_CHAR:
		switch( wParam )
		{
		case 'n':
		case 'N':
			m_bShowNormals = !m_bShowNormals;
			break;

		case 's':
		case 'S':
			m_bShowSurface = !m_bShowSurface;
			break;

		case 'c':
		case 'C':
			m_bShowControl = !m_bShowControl;
			break;
		}
		break;
	}

	return CD3DApp::MsgProc(hWnd, uMsg, wParam, lParam);
}





//-----------------------------------------------------------------------------
// CubicBezier()
//-----------------------------------------------------------------------------
void CTriPatch::CubicBezier( CVertex P[10] )
{
	// Definitions for triangular bezier patches
	//-------------------------------------------------------------------------
		// Triangle corners
		// t0 = Lower left
		// t1 = Top
		// t2 = Lower right

		// Barycentric coordinates (2D)
		// Area = (t1.x-t0.x)*(t2.y-t0.y) - (t1.y-t0.y)*(t2.x-t0.x)
		// lambda0(u) = ((t1.x-u.x)*(t2.y-u.y) - (t1.y-u.y)*(t2.x-u.x))/Area
		// lambda1(u) = ((t2.x-u.x)*(t0.y-u.y) - (t2.y-u.y)*(t0.x-u.x))/Area
		// lambda2(u) = ((t0.x-u.x)*(t1.y-u.y) - (t0.y-u.y)*(t1.x-u.x))/Area

		// Bezier points (3D)
		// bijk = (i*t0 + j*t1 + k*t2)/n
		// n = degree of bezier patch
		// i + j + k = n
		//
		//         b002
		//     b101    b011       
		// b200    b110    b020   

		// Bernstein Polynomials
		// B000(u) = 1
		// Bijk(u) = lambda0(u)*Bi-1,j,k(u) + 
	    //           lambda1(u)*Bi,j-1,k(u) + 
	    //           lambda2(u)*Bi,j,k-1(u)
		// (Negative indices are set to zero)

		// Evaluation
		// F(u) = Sum(Bijk(u)bijk, i+j+k=n)

		//-----

		// de Casteljau's Algorithm (the pyramid)
		// 
		//         b0_002
		//     b0_101  b0_011            b1_001
		// b0_200  b0_110  b0_020    b1_100  b1_010    b2_000
		
		// de C (evaluation)
		// F(u) = b2_000
		// b0_ijk = bijk
		// bl_ijk = lambda0(u)*bl-1_i+1,j,k + 
	    //          lambda1(u)*bl-1_i,j+1,k + 
	    //          lambda2(u)*bl-1_i,j,k+1

		// de C (tangent plane)
		// the tangent plane is spanned by
		// b1_100, b1_010, b1_001
	//-------------------------------------------------------------------------

	// We'll use de Casteljau's Algorithm since it's easier to understand

	// The control points are ordered like this
	//
	//    3            003
	//   2 6         102 012
	//  1 5 8      201 111 021
	// 0 4 7 9   300 210 120 030

	// Compute the vertices
	int Index = 0;
	for( int n = 0; n < m_nNumDivs; n++ )
	{
		for( int i = 0; i < m_nNumDivs-n; i++ )
		{
			// The barycentric coordinates
			float lambda1 = float(i)/(m_nNumDivs-1);
			float lambda2 = float(n)/(m_nNumDivs-1);
			float lambda0 = 1-lambda1-lambda2;

			// Move up the pyramid as our wanted surface coordinate is at the top
			//
			//     002
			//   101 011
			// 200 110 020
			CVertex b1_200, b1_110, b1_020, b1_101, b1_011, b1_002;
			b1_200.v = lambda0*P[0].v + lambda1*P[1].v + lambda2*P[4].v;
			b1_110.v = lambda0*P[1].v + lambda1*P[2].v + lambda2*P[5].v;
			b1_020.v = lambda0*P[2].v + lambda1*P[3].v + lambda2*P[6].v;
			b1_101.v = lambda0*P[4].v + lambda1*P[5].v + lambda2*P[7].v;
			b1_011.v = lambda0*P[5].v + lambda1*P[6].v + lambda2*P[8].v;
			b1_002.v = lambda0*P[7].v + lambda1*P[8].v + lambda2*P[9].v;

			//   001
			// 100 010
			CVertex b2_100, b2_010, b2_001;
			b2_100.v = lambda0*b1_200.v + lambda1*b1_110.v + lambda2*b1_101.v;
			b2_010.v = lambda0*b1_110.v + lambda1*b1_020.v + lambda2*b1_011.v;
			b2_001.v = lambda0*b1_101.v + lambda1*b1_011.v + lambda2*b1_002.v;

			m_pVertices[Index].v = lambda0*b2_100.v + 
				                   lambda1*b2_010.v + 
								   lambda2*b2_001.v;
			m_pVertices[Index].n = Normalize(CrossProduct(b2_010.v - b2_100.v, 
				                                          b2_001.v - b2_100.v));
			m_pVertices[Index].c = D3DRGBA(1,1,1,1);
			Index++;
		}
	}	
}

